----------------------------------------------------------------------------------
----------------------------------------------------------------------------------
-- Filename     HYSyncGen.vhd
-- Title        CRT controller
--
-- Company      Design Gateway Co., Ltd.
-- Project      DRID
-- PJ No.       DS0502
-- Syntax       VHDL
-- Note         This module generates Hsync, Vsync, Blanking, RGB to VGA
--
-- Version      0.01
-- Author       U.Patheera
-- Date         2005/03/31
-- Remark       New Creation
----------------------------------------------------------------------------------
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

Entity HYSyncGen Is
	Generic (
		Hres		: std_logic_vector(10 downto 0 ):= "01100100000";-- Horizontal resolution (800 pixels/line)
		Vres		: std_logic_vector(10 downto 0 ):= "01000001101";-- Vertical resolution (525 lines/frame)
		HsyncWidth	: std_logic_vector( 6 downto 0 ):= "1100000";	 -- Horizontal sync width (96 pixels)
		VsyncWidth	: std_logic_vector( 6 downto 0 ):= "0000010";	 -- Vertical sync width (2 lines)
		Hbackporch	: std_logic_vector( 6 downto 0 ):= "0101101";	 -- Horizontal back porch 45 pixels)
		Hactive		: std_logic_vector(10 downto 0 ):= "01010000000";-- Horizontal active pixel (640 pixels/line)
		Vbackporch	: std_logic_vector( 6 downto 0 ):= "0011110";	 -- Vertical back porch (30 lines)
		Vactive		: std_logic_vector(10 downto 0 ):= "00111100000";-- Vertical active line (480 lines/frame)
		HVpolar		: std_logic_vector( 2 downto 0 ):= "110";		 -- Hsync/Vsync/Blank polarity (Sync Active Low)

		MAXROW		: std_logic_vector(13 downto 0 ):= "00000000000100";--(N-1 rows)
		MAXLINE 	: std_logic_vector( 8 downto 0 ):= "000000001"		--(N-1 lines)
	);
    Port (
		Clk100      		: in    std_logic;                      
		Clk25				: in	std_logic;
		RstB        		: in    std_logic;                      
	
		Busy				: in	std_logic;
		CmdReq				: out 	std_logic;
		Cmd					: out	std_logic;
		RdAddr				: out	std_logic_vector( 14 downto 0 );
		WrCnt				: in	std_logic_vector( 10 downto 9 );

		AlmostEmpty			: in	std_logic;
		Empty				: in	std_logic;
		RdEn				: out	std_logic;
		RdData				: in	std_logic_vector( 31 downto 0 );
		
		PageSel				: in	std_logic;
		
		Hsync_out   		: out   std_logic;                      
		Vsync_out   		: out   std_logic;                      
		Blank_out   		: out   std_logic;                      
		Red_out				: out	std_logic_vector( 7 downto 0 );	
		Green_out			: out	std_logic_vector( 7 downto 0 );	
		Blue_out			: out	std_logic_vector( 7 downto 0 )
    );
End Entity HYSyncGen;

Architecture rtl Of HYSyncGen Is

--------------------------------------------------------------------------------
-- Constant Declaration
--------------------------------------------------------------------------------
	type state is (
					stIdle,
					stCmdRead,
					stReadData,
					stWait,
					stEnd
				);
				
	signal	rState			: state;

----------------------------------------------------------------------------------
-- Signal declaration
----------------------------------------------------------------------------------
	signal	rCnt_Hres 		: std_logic_vector(10 downto 0 ); 	-- Horizontal resolution counter
	signal	rCnt_Vres 		: std_logic_vector(10 downto 0 ); 	-- Vertical resolution counter
	signal	rCnt_HsyncW 	: std_logic_vector( 6 downto 0 );	-- Hor-sync width
	signal	rCnt_VsyncW 	: std_logic_vector( 6 downto 0 );	-- Ver-sync width
	signal	rCnt_Hbackp 	: std_logic_vector( 6 downto 0 );	-- H-backporch width
	signal	rCnt_Vbackp 	: std_logic_vector( 6 downto 0 );	-- V-backporch width
	signal	rCnt_H_Enable 	: std_logic_vector(10 downto 0 );	-- Hor-Pixel enable counter
	signal	rCnt_V_Enable 	: std_logic_vector(10 downto 0 );	-- Ver-Pixel enable counter

	signal	rHsync 			: std_logic;						-- H-sync signal
	signal	rVsync 			: std_logic;						-- V-sync signal
	signal	rHbackp 		: std_logic;						-- H-backporch signal
	signal	rVbackp 		: std_logic;						-- V-backporch signal
	signal	rH_Enable 		: std_logic;						-- H-Enable signal
	signal	rV_Enable 		: std_logic;						-- V-Enable signal

	signal	rHsync_out 		: std_logic;						-- H-sync with selected polarity
	signal	rVsync_out 		: std_logic;						-- V-sync with selected polarity
	signal	rBlank_out 		: std_logic;						-- Blanking with selected polarity

	signal	rRed_out		: std_logic_vector( 7 downto 0 );	-- Red output
	signal	rGreen_out		: std_logic_vector( 7 downto 0 );	-- Green output
	signal	rBlue_out		: std_logic_vector( 7 downto 0 );	-- Blue output

	signal	dlyHsync_out 	: std_logic_vector( 6 downto 0 );	-- Delay of H-sync
	signal	dlyVsync_out 	: std_logic_vector( 8 downto 0 );	-- Delay of V-sync
	signal	dlyBlank_out 	: std_logic_vector( 6 downto 0 );	-- Delay of Blanking											    		   

	signal	rCmdReq			: std_logic;
	signal	rRdAddr			: std_logic_vector(14 downto 0 );
	signal	rRdEn			: std_logic;
	signal	rPageSel		: std_logic;
	
	signal	rCnt80			: std_logic_vector( 6 downto 0 );
	signal	rCnt8			: std_logic_vector( 2 downto 0 );
	
Begin

----------------------------------------------------------------------------------
-- Output assignment
----------------------------------------------------------------------------------
	HSYNC_out	<= dlyHsync_out(2); 
	VSYNC_out	<= dlyVsync_out(2); 
	BLANK_out	<= dlyBlank_out(2); 
	Red_out	 	<= rRed_out;
	Green_out	<= rGreen_out;
	Blue_out 	<= rBlue_out;
	
	CmdReq		<= rCmdReq;
	Cmd			<= '1';
	RdAddr		<= rRdAddr;
	RdEn		<= rRdEn;	

----------------------------------------------------------------------------------
-- DFF part
----------------------------------------------------------------------------------

--********************* Horizontal Sync Generate *************************	

	-- Cnt_Hres signal used to count Hres time
	u_Cnt_Hres : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rCnt_Hres( 10 downto 0 ) <= Hres;
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				rCnt_Hres <= Hres( 10 downto 0 );
			else
				rCnt_Hres <= rCnt_Hres-1;
			end if;
		end if;
	End Process u_Cnt_Hres;	

	-- Horizontal Sync will be HIGH when finish active pixel, detected from Hres. 
	-- It will be reset to LOW after Cnt_HsyncW = 1.
	u_Hsync : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rHsync <= '0';
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				rHsync <= '1';
			elsif ( rCnt_HsyncW = 1 ) then  
				rHsync <= '0';
			end if;
		end if;
	End Process u_Hsync;
	
	-- Cnt_HsyncW uses to count Hsync time	
	u_Cnt_HsyncW : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rCnt_HsyncW ( 6 downto 0 ) <= HsyncWidth;
		elsif ( rising_edge(Clk25) ) then
			if ( rHsync = '1' ) then
				rCnt_HsyncW <= rCnt_HsyncW-1;
			else
				rCnt_HsyncW <= HsyncWidth;
			end if;  	
		end if;
	End Process u_Cnt_HsyncW;
	
	-- Hbackp will be HIGH when that time is Horizontal back porch and will reset to LOW
	-- when Cnt_Hbackp = 1.					
	u_Hbackp : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rHbackp <= '0';
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_HsyncW = 1 ) then
				rHbackp <= '1';
			elsif ( rCnt_Hbackp = 1 ) then
				rHbackp <= '0';
			end if;
		end if;
	End Process u_Hbackp;
   
	-- Cnt_Hbackp uses to count Hbackporch signal	
	u_Cnt_Hbackp : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rCnt_Hbackp ( 6 downto 0 ) <= Hbackporch;
		elsif ( rising_edge(Clk25) ) then
			if ( rHbackp = '1' ) then
				rCnt_Hbackp <= rCnt_Hbackp-1;
			else
				rCnt_Hbackp <= Hbackporch;
			end if;
		end if;	
	End Process u_Cnt_Hbackp;		
	
	-- H_Enable = 1 when that pixel is active pixel and will reset to 0 when Cnt_H_Enable = 1.
	u_H_Enable : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rH_Enable <= '0';
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hbackp = 1 ) then
				rH_enable <= '1';
			elsif ( rCnt_H_Enable = 1 ) then
				rH_enable <= '0';
			end if;
		end if;
	End Process u_H_Enable;
	
	-- Cnt_H_Enable uses to count H_Enable signal
	u_Cnt_H_Enable : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rCnt_H_Enable <= Hactive;
		elsif ( rising_edge(Clk25) ) then
			if ( rH_enable = '1' ) then
				rCnt_H_Enable <= rCnt_H_Enable-1;
			else
				rCnt_H_Enable <= Hactive;
			end if;
		end if;
	End Process u_Cnt_H_Enable;


--********************* Vertical Sync Generate *************************

	-- Cnt_Hres signal used to count Vres time
	u_Cnt_Vres : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rCnt_Vres( 10 downto 0 ) <= Vres;
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				if ( rCnt_Vres = 1 ) then
					rCnt_Vres <= Vres( 10 downto 0 );
				else
					rCnt_Vres <= rCnt_Vres-1;
				end if;
			end if;
		end if;
	End Process u_Cnt_Vres;	

	-- Vorizontal Sync will be HIGH when finish active pixel, detected from Vres. 
	-- It will be reset to LOW after Cnt_VsyncW = 1.
	u_Vsync : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rVsync <= '0';
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				if ( rCnt_Vres = 1 ) then
					rVsync <= '1';
				elsif ( rCnt_VsyncW = 1 ) then  
					rVsync <= '0';
				end if;
			end if;
		end if;
	End Process u_Vsync;
	
	-- Cnt_VsyncW uses to count Vsync time
	u_Cnt_VsyncW : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rCnt_VsyncW ( 6 downto 0 ) <= VsyncWidth;
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				if ( rVsync = '1' ) then
					rCnt_VsyncW <= rCnt_VsyncW-1;
				else
					rCnt_VsyncW <= VsyncWidth;
				end if;  	
			end if;
		end if;
	End Process u_Cnt_VsyncW;
	
	-- Vbackp will be HIGH when that time is Vertical back porch and will reset to LOW
	-- when Cnt_Vbackp = 1.					
	u_Vbackp : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rVbackp <= '0';
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				if ( rCnt_VsyncW = 1 ) then
					rVbackp <= '1';
				elsif ( rCnt_Vbackp = 1 ) then
					rVbackp <= '0';
				end if;
			end if;
		end if;
	End Process u_Vbackp;
	
	-- Cnt_Vbackp uses to count Vbackporch signal	
	u_Cnt_Vbackp : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rCnt_Vbackp ( 6 downto 0 ) <= Vbackporch;
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				if ( rVbackp = '1' ) then
					rCnt_Vbackp <= rCnt_Vbackp-1;
				else
					rCnt_Vbackp <= Vbackporch;
				end if;
			end if;
		end if;	
	End Process u_Cnt_Vbackp;		

	-- V_Enable = 1 when that line is active line and will reset to 0 when Cnt_V_Enable = 1.
	u_V_Enable : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rV_Enable <= '0';
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				if ( rCnt_Vbackp = 1 ) then
					rV_enable <= '1';
				elsif ( rCnt_V_Enable = 1 ) then
					rV_enable <= '0';
				end if;
			end if;
		end if;
	End Process u_V_Enable;

	-- Cnt_V_Enable uses to count V_Enable signal
	u_Cnt_V_Enable : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rCnt_V_Enable <= Vactive;
		elsif ( rising_edge(Clk25) ) then
			if ( rCnt_Hres = 1 ) then
				if ( rV_enable = '1' ) then
					rCnt_V_Enable <= rCnt_V_Enable-1;
				else
					rCnt_V_Enable <= Vactive;
				end if;
			end if;
		end if;
	End Process u_Cnt_V_Enable;

	-- If HVpolar(2) = '1', Hsync_out = '0' when Horizontal sync time.
	-- If HVpolar(2) = '0', Hsync_out = '1' when Horizontal sync time.
	u_Hsync_out : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rHsync_out <= '1';
		elsif ( rising_edge(Clk25) ) then
			rHsync_out	<= ( rHsync xor HVpolar(2) );
		end if;
	End Process u_Hsync_out;
	
	-- If HVpolar(1) = '1', Vsync_out = '0' when Vertical sync time.
	-- If HVpolar(1) = '0', Vsync_out = '1' when Vertical sync time.
	u_Vsync_out : Process (RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rVsync_out <= '1';
		elsif ( rising_edge(Clk25) ) then
			rVsync_out <= ( rVsync xor HVpolar(1) );
		end if;
	End Process u_Vsync_out;
	
	-- If HVpolar(0) = '1', Blank_out = '0' when blanking time.
	-- If HVpolar(0) = '0', Blank_out = '1' when blanking time.
	u_Blank_out : Process ( RstB, Clk25) Is
	Begin
		if ( RstB = '0' ) then
			rBlank_out <= '1';
		elsif ( rising_edge(Clk25) ) then
			rBlank_out	<= ( not ( rH_Enable and rV_Enable ) ) xor HVpolar(0);
		end if;
	End Process u_Blank_out;

	-- dly uses for delay sync output bye one clock.
	u_dlyHVB : Process ( Clk25 ) Is
	Begin
		if ( rising_edge(Clk25) ) then
			dlyHsync_out( 6 downto 0 ) <= dlyHsync_out(5 downto 0) & rHsync_out;
			dlyVsync_out( 8 downto 0 ) <= dlyVsync_out(7 downto 0) & rVsync_out;
			dlyBlank_out( 6 downto 0 ) <= dlyBlank_out(5 downto 0) & rBlank_out;
		end if;
	End Process u_dlyHVB;

	
	----------------------------------------------------------------------------
	-- State Machine
	----------------------------------------------------------------------------

	-- Command Path ------------------------------------------------------------
	u_rState : Process (RstB, Clk100) Is
	Begin
		if ( RstB='0' ) then
			rState	<= stIdle;
		elsif ( rising_edge(Clk100) ) then
			case (rState) is
			
				when stIdle		=>
					if ( Busy='0' and WrCnt(10)='0' ) then
						rState	<= stCmdRead;
					else
						rState	<= stIdle;
					end if;
					
				when stCmdRead	=>
					if ( Busy='1' ) then
						rState	<= stReadData;
					else
						rState	<= stCmdRead;
					end if;
					
				when stReadData =>
					if ( Busy='0' ) then
						if ( rRdAddr(13 downto 0)=MAXROW ) then
							rState	<= stEnd;
						else
							rState	<= stIdle;
						end if;
					else
						rState	<= stReadData;
					end if;
				
				when stEnd		=>
					rState	<= stIdle;
				
				when others		=>
					rState	<= stIdle;
					
			end case;
		end if;
	End Process u_rState;
	
	u_rCmdReq : Process (RstB, Clk100) Is
	Begin
		if ( RstB='0' ) then
			rCmdReq	<= '0';
		elsif ( rising_edge(Clk100) ) then
			if ( Busy='1' ) then
				rCmdReq	<= '0';
			elsif ( rState=stCmdRead ) then
				rCmdReq	<= '1';
			else
				rCmdReq	<= rCmdReq;
			end if;
		end if;
	End Process u_rCmdReq;
	
	u_rRdAddr : Process (RstB, Clk100) Is
	Begin
		if ( RstB='0' ) then
			rRdAddr	<= (others => '0');
		elsif ( rising_edge(Clk100) ) then
			if ( rState=stReadData and Busy='0' ) then
				if ( rRdAddr(13 downto 0)=MAXROW ) then
					rRdAddr(13 downto 0)	<= (others => '0');
				else
					rRdAddr(13 downto 0)	<= rRdAddr(13 downto 0) + 1;
				end if;
			else
				rRdAddr(13 downto 0)	<= rRdAddr(13 downto 0);
			end if;
			
			if ( rState=stEnd ) then
				rRdAddr(14)		<= PageSel;
			else
				rRdAddr(14)		<= rRdAddr(14);
			end if;
		end if;
	End Process u_rRdAddr;
	
	-- Data Path ---------------------------------------------------------------
	u_rRdEn : Process (RstB, Clk25) Is
	Begin
		if ( RstB='0' ) then
			rRdEn	<= '0';
		elsif ( rising_edge(Clk25) ) then
			if ( AlmostEmpty='1' ) then
				rRdEn	<= '0';
			else
				rRdEn	<= not rBlank_out;
			end if;
		end if;
	End Process u_rRdEn;

	u_RGB : Process (RstB, Clk25) Is
	Begin
		if ( RstB='0' ) then
			rRed_out	<= x"00";	
			rGreen_out	<= x"00";
			rBlue_out	<= x"00";
		elsif ( rising_edge(Clk25) ) then
			rRed_out	<= RdData(23 downto 16);
			rGreen_out	<= RdData(15 downto 8 );
			rBlue_out	<= RdData( 7 downto 0 );
		end if;
	End Process u_RGB;
	
End Architecture rtl;